__all__ = [
    "BaseDataConnection"
]

#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import io
import json
import os

from abc import ABC, abstractmethod
from typing import Union, Tuple, TYPE_CHECKING
from warnings import warn

import ibm_watson_machine_learning._wrappers.requests as requests
from pandas import concat, DataFrame, read_csv

from ibm_watson_machine_learning.utils.autoai.utils import try_load_dataset, try_load_tar_gz, try_import_autoai_libs
from ibm_watson_machine_learning.utils.autoai.enums import DataConnectionTypes
from ibm_watson_machine_learning.utils.autoai.errors import CannotReadSavedRemoteDataBeforeFit, WrongAssetType
from ibm_watson_machine_learning.wml_client_error import ApiRequestFailure, WMLClientError, DataStreamError
from ibm_watson_machine_learning.utils.utils import is_lib_installed
from ibm_watson_machine_learning.data_loaders.datasets.experiment import DEFAULT_SAMPLING_TYPE, DEFAULT_SAMPLE_SIZE_LIMIT

if TYPE_CHECKING:
    from ibm_boto3 import resource


class BaseDataConnection(ABC):
    """
    Base class for  DataConnection.
    """

    def __init__(self):

        self.type = None
        self.connection = None
        self.location = None
        self.auto_pipeline_params = None
        self._wml_client = None
        self._run_id = None
        self._obm = None
        self._obm_cos_path = None
        self.id = None
        self._datasource_type = None

    @abstractmethod
    def _to_dict(self) -> dict:
        """Convert DataConnection object to dictionary representation."""
        pass

    @classmethod
    @abstractmethod
    def _from_dict(cls, _dict: dict) -> 'BaseDataConnection':
        """Create a DataConnection object from dictionary."""
        pass

    @abstractmethod
    def read(self, with_holdout_split: bool = False) -> Union['DataFrame', Tuple['DataFrame', 'DataFrame']]:
        """Download dataset stored in remote data storage."""
        pass

    @abstractmethod
    def write(self, data: Union[str, 'DataFrame'], remote_name: str) -> None:
        """Upload file to a remote data storage."""
        pass

    def _fill_experiment_parameters(self, prediction_type: str, prediction_column: str, holdout_size: float,
                                    csv_separator: str = ',', excel_sheet: Union[str, int] = None,
                                    encoding: str = 'utf-8') -> None:
        """
        To be able to recreate a holdout split, this method need to be called.
        """
        self.auto_pipeline_params = {
            'prediction_type': prediction_type,
            'prediction_column': prediction_column,
            'holdout_size': holdout_size,
            'csv_separator': csv_separator,
            'excel_sheet': excel_sheet,
            'encoding': encoding
        }

    def _download_obm_data_from_cos(self, cos_client: 'resource') -> 'DataFrame':
        """Download preprocessed OBM data. COS version."""

        if '//' in self._obm_cos_path:  # sometimes there is an additional slash, need to replace it
            self._obm_cos_path = self._obm_cos_path.replace('//', '/')

        # note: fetch all OBM file part names
        cos_summary = cos_client.Bucket(self.location.bucket).objects.filter(Prefix=self._obm_cos_path)
        file_names = [file_name.key for file_name in cos_summary]

        # note: if path does not exist, try to find in different one
        if not file_names:
            cos_summary = cos_client.Bucket(self.location.bucket).objects.filter(
                Prefix=self._obm_cos_path.split('./')[-1])
            file_names = [file_name.key for file_name in cos_summary]
            # --- end note
        # --- end note

        # TODO: this can be done simultaneously (multithreading / multiprocessing)
        # note: download all data parts and concatenate them into one output
        parts = []
        for file_name in file_names:
            file = cos_client.Object(self.location.bucket, file_name).get()
            buffer = io.BytesIO(file['Body'].read())
            parts.append(try_load_dataset(buffer=buffer))

        data = concat(parts)
        # --- end note
        return data

    def _download_obm_json(self) -> dict:
        """Download obm.json log."""
        if self._obm:
            if self._obm_cos_path:
                path = self._obm_cos_path.rsplit('features', 1)[0] + 'obm.json'
            else:
                path = f"{self.location.path}/{self._run_id}/data/obm/obm.json"
            data = self._download_json_file(path)
            return data
        else:
            raise Exception("OBM function called when not OBM scenario.")

    def _download_json_file(self, path) -> dict:
        """Download json file."""
        json_content = {}

        if '//' in path:  # sometimes there is an additional slash, need to replace it
            path = path.replace('//', '/')

        # TODO: remove S3 implementation
        if self.type == DataConnectionTypes.S3:
            warn(message="S3 DataConnection is deprecated! Please use data_asset_id instead.")

            cos_client = self._init_cos_client()

            try:
                file = cos_client.Object(self.location.bucket, path).get()
                content = file["Body"].read()
                json_content = json.loads(content.decode('utf-8'))
            except Exception as cos_access_exception:
                raise ConnectionError(
                    f"Unable to access data object in cloud object storage with credentials supplied. "
                    f"Error: {cos_access_exception}")
        elif self.type == DataConnectionTypes.FS:
            json_url = self._wml_client.service_instance._href_definitions.get_wsd_model_attachment_href() + f"/auto_ml/{path.split('/auto_ml/')[-1]}"

            with requests.get(json_url, params=self._wml_client._params(), headers=self._wml_client._get_headers(),
                              stream=True) as file_response:
                if file_response.status_code != 200:
                    raise ApiRequestFailure(u'Failure during {}.'.format("downloading json"), file_response)

                downloaded_asset = file_response.content
                buffer = io.BytesIO(downloaded_asset)
                json_content = json.loads(buffer.getvalue().decode('utf-8'))

        elif self.type == DataConnectionTypes.CA or self.type == DataConnectionTypes.CN:
            if self._check_if_connection_asset_is_s3():
                cos_client = self._init_cos_client()

                try:
                    file = cos_client.Object(self.location.bucket, path).get()
                    content = file["Body"].read()
                    json_content = json.loads(content.decode('utf-8'))

                except Exception as cos_access_exception:
                    raise ConnectionError(
                        f"Unable to access data object in cloud object storage with credentials supplied. "
                        f"Error: {cos_access_exception}")

            else:
                raise NotImplementedError(f"Unsupported connection type: {self.type}. "
                                          f"Datasource type is not supported. "
                                          f"Supported type is: bluemixcloudobjectstorage")

        else:
            raise NotImplementedError(f"Unsupported connection type: {self.type}")

        return json_content

    def _get_attachemnt_details(self, data_asset_id: str):
        response = requests.get(self._wml_client.service_instance._href_definitions.get_data_asset_href(data_asset_id),
                                params=self._wml_client._params(),
                                headers=self._wml_client._get_headers())
        data_asset_details = self._wml_client.data_assets._handle_response(200, u'Cannot get data asset details', response)

        attachments_data_asset_details = data_asset_details.get('attachments', [{}])[0]
        # note: Return attachment details from data asset details if it is not a connected data asset:
        if attachments_data_asset_details and attachments_data_asset_details.get('connection_id') is None:
            return attachments_data_asset_details
        else:
            attachment_id = attachments_data_asset_details.get('id')
            if not self._wml_client.ICP and not self._wml_client.WSD:
                response = requests.get(
                    self._wml_client.service_instance._href_definitions.get_attachment_href(data_asset_id, attachment_id),
                    params=self._wml_client._params(),
                    headers=self._wml_client._get_headers())
            else:
                response = requests.get(
                    self._wml_client.service_instance._href_definitions.get_attachment_href(data_asset_id, attachment_id),
                    params=self._wml_client._params(),
                    headers=self._wml_client._get_headers(),
                    verify=False)

            self._wml_client.data_assets._handle_response(200, u'Cannot get attachment details', response)

            return response.json()

    def _check_if_connection_asset_is_s3(self) -> bool:
        try:
            # fast return true if the connection is an instance of S3Connection and non s3 type, or its attribute
            # "cos_type" takes True
            from .connections import S3Connection
            if (isinstance(self.connection, S3Connection) and self.type != 's3' and self.connection.to_dict()) or (
                    self.connection is not None and getattr(self.connection, 'cos_type', False)):
                return True

            if self.type == 'data_asset':
                if self.connection is not None and hasattr(self.connection, 'href'):
                    items = self.connection.href.split('/')

                else:
                    items = self.location.href.split('/')
                _id = items[-1].split('?')[0]

                if self._wml_client is not None:
                    attachment_details = self._get_attachemnt_details(_id)
                    if 'connection_id' not in attachment_details:
                        if 'handle' in attachment_details and 'bucket' in attachment_details['handle']:
                            connection_details = self._create_conn_details_for_container()
                            attachment_details['connection_path'] = '/' + attachment_details['handle']['bucket'] + '/' + attachment_details['handle']['key']
                        elif 'url' in attachment_details and 'name' in attachment_details and 'bucket' in attachment_details:
                            connection_details = self._create_conn_details_for_container()
                            attachment_details['connection_path'] = '/' + attachment_details['bucket']['bucket_name'] + '/' + attachment_details['name']
                        else:
                            return False

                    else:
                        connection_details = self._wml_client.connections.get_details(
                            attachment_details['connection_id'])

                else:
                    try:
                        from project_lib import Project
                        project = Project.access()

                        # note: Check if asset is located directly in the project files
                        #       If yes it is not a connected data.
                        #       [Prevents unnecessary logging].
                        if any(file['asset_id'] == _id for file in project.get_files()):
                            return False
                        # --- end note

                        try:
                            connection_details = project.get_connected_data(_id)

                        except RuntimeError:  # when data asset is normal not s3 or database
                            return False

                        attachment_details = {'connection_path': connection_details.get('datapath')}

                    except ModuleNotFoundError:
                        raise NotImplementedError(f"This functionality can be run only on Watson Studio.")

            elif self.type == 'connection_asset':
                if self._wml_client is not None:
                    connection_details = self._wml_client.connections.get_details(self.connection.id)

                else:
                    try:
                        from project_lib import Project
                        project = Project.access()
                        connection_details = project.get_connection(self.connection.id)

                    except ModuleNotFoundError:
                        raise NotImplementedError(f"This functionality can be run only on Watson Studio.")

            elif self.type == 'container':
                connection_details = self._create_conn_details_for_container()

            elif self.type == 's3' or self.type == 'fs':
                return False

            else:
                raise WrongAssetType(asset_type=self.type)

            # Note: Check with project libs if connection points to S3
            if self._wml_client is not None:
                datasource_type = connection_details['entity']['datasource_type']
                self._datasource_type = datasource_type
                datasource_type_id_ibm_cos = self._wml_client.connections.get_datasource_type_uid_by_name(
                    'bluemixcloudobjectstorage')
                datasource_type_id_aws_cos = self._wml_client.connections.get_datasource_type_uid_by_name(
                    'cloudobjectstorage')

                if (datasource_type == datasource_type_id_ibm_cos or datasource_type == datasource_type_id_aws_cos or
                        datasource_type == 'bluemixcloudobjectstorage' or datasource_type == 'cloudobjectstorage'):
                    cos_type = True

                else:
                    cos_type = False

            elif self.type == 'container':
                cos_type = True

            elif 'url' in connection_details:
                cos_type = True
                connection_details['entity'] = {'properties': connection_details}

            else:
                cos_type = False
            # --- end note

            if cos_type:
                if 'cos_hmac_keys' in str(connection_details['entity']['properties']):
                    creds = json.loads(connection_details['entity']['properties']['credentials'])
                    connection_details['entity']['properties']['access_key'] = creds['cos_hmac_keys']['access_key_id']
                    connection_details['entity']['properties']['secret_key'] = creds['cos_hmac_keys'][
                        'secret_access_key']

                if self.connection is None:
                    from .connections import S3Connection
                    if ('api_key' in str(connection_details['entity']['properties']) and
                            'iam_url' in str(connection_details['entity']['properties'])):
                        self.connection = S3Connection(
                            api_key=connection_details['entity']['properties'].get('api_key'),
                            auth_endpoint=connection_details['entity']['properties'].get('iam_url'),
                            endpoint_url=connection_details['entity']['properties'].get('url'),
                            resource_instance_id=connection_details['entity']['properties'].get('resource_instance_id'),
                            _internal_use=True
                        )

                    else:
                        self.connection = S3Connection(
                            endpoint_url=connection_details['entity']['properties'].get('url'),
                            access_key_id=connection_details['entity']['properties'].get('access_key'),
                            secret_access_key=connection_details['entity']['properties'].get('secret_key'),
                            _internal_use=True
                        )

                else:
                    if ('api_key' in str(connection_details['entity']['properties']) and
                            'iam_url' in str(connection_details['entity']['properties'])):
                        self.connection.api_key = connection_details['entity']['properties'].get('api_key')
                        self.connection.auth_endpoint = connection_details['entity']['properties'].get('iam_url')
                        self.connection.endpoint_url = connection_details['entity']['properties'].get('url')
                        self.connection.resource_instance_id = connection_details['entity']['properties'].get('resource_instance_id')

                    else:
                        self.connection.secret_access_key = connection_details['entity']['properties'].get('secret_key')
                        self.connection.access_key_id = connection_details['entity']['properties'].get('access_key')
                        self.connection.endpoint_url = connection_details['entity']['properties'].get('url')

                if self.type == 'container':
                    self.location.bucket = connection_details['entity']['properties']['bucket_name']

                if self.type == 'data_asset':
                    try:
                        # Workaround for connected data assets promoted from catalog
                        if not attachment_details['connection_path'].startswith('/'):
                            attachment_details['connection_path'] = '/' + attachment_details['connection_path']
                        self.location.bucket = attachment_details['connection_path'].split('/')[1]

                        # Remove excel sheet name from connection_path if data asset is excel file:
                        if '.xlsx/' in attachment_details['connection_path'] or '.xls/' in attachment_details['connection_path']:
                            excel_sheet = attachment_details['connection_path'].split('/')[-1]
                            # note: set recognized excel sheet in params
                            if not self.auto_pipeline_params.get('excel_sheet'):
                                self.auto_pipeline_params['excel_sheet'] = excel_sheet
                            # end note
                            self.location.path = '/'.join(attachment_details['connection_path'].split('/')[2:-1])
                        else:
                            self.location.path = '/'.join(attachment_details['connection_path'].split('/')[2:])

                    except IndexError:
                        self.location.bucket = connection_details['entity']['properties']['bucket']
                        self.location.path = attachment_details['connection_path']

                    self.location.id = None
                self.connection.cos_type = True
                return True

            else:
                return False

        except Exception as e:
            if os.environ.get('USER_ACCESS_TOKEN') is None and os.environ.get('RUNTIME_ENV_ACCESS_TOKEN_FILE') is None:
                raise e

            else:
                return False  # if we are in WS, ignore this check even if there was some error

    def _is_data_asset_normal(self) -> bool:
        """Returns true if data asset is normal data asset - not connected data asset"""
        try:
            if self.type == 'data_asset':
                if self.connection is not None and hasattr(self.connection, 'href'):
                    items = self.connection.href.split('/')
                else:
                    items = self.location.href.split('/')
                _id = items[-1].split('?')[0]

                if self._wml_client is not None:
                    attachment_details = self._get_attachemnt_details(_id)
                    return bool('connection_id' not in attachment_details)

                else:
                    try:
                        from project_lib import Project
                        project = Project.access()

                        # note: Check if asset is located directly in the project files
                        #       If yes it is not a connected data.
                        #       [Prevents unnecessary logging].
                        return any(file['asset_id'] == _id for file in project.get_files())
                        # --- end note

                    except ModuleNotFoundError:
                        raise NotImplementedError(f"This functionality can be run only on Watson Studio.")

        except Exception as e:
            if os.environ.get('USER_ACCESS_TOKEN') is None and os.environ.get('RUNTIME_ENV_ACCESS_TOKEN_FILE') is None:
                raise e
            else:
                return False  # if we are in WS, ignore this check even if there was some error

    def _is_data_asset_nfs(self):
        if self.type == 'data_asset':
            if self.connection is not None and hasattr(self.connection, 'href'):
                items = self.connection.href.split('/')
            else:
                items = self.location.href.split('/')
            _id = items[-1].split('?')[0]

            if self._wml_client is not None:
                attachment_details = self._get_attachemnt_details(_id)

                return ('connection_id' in attachment_details and attachment_details.get(
                    'datasource_type') == self._wml_client.connections.get_datasource_type_uid_by_name('volumes'))
        return False


    def _download_indices_from_cos(self, cos_client: 'resource', location_path) -> 'DataFrame':
        """Download indices for this connection. COS version"""

        try:
            file = cos_client.Object(self.location.bucket,
                                     location_path).get()
        except:
            file = list(cos_client.Bucket(self.location.bucket).objects.filter(
                Prefix=location_path))[0].get()

        buffer = io.BytesIO(file['Body'].read())

        if '.csv' in location_path:
            file_name = 'indices.csv'
            with open(file_name, 'wb') as out:
                out.write(buffer.read())

            data = read_csv(file_name, sep=self.auto_pipeline_params.get('csv_separator', ','),
                            encoding=self.auto_pipeline_params.get('encoding', 'utf-8'))

        else:
            data = try_load_tar_gz(buffer=buffer,
                                   separator=self.auto_pipeline_params.get('csv_separator', ','),
                                   encoding=self.auto_pipeline_params.get('encoding', 'utf-8')
                                   )

        return data

    def _download_training_data_from_data_asset_storage(self) -> 'DataFrame':
        """Download training data for this connection. Data Storage."""

        if self._wml_client is not None:
            # note: as we need to load a data into the memory,
            # we are using pure requests and helpers from the WML client
            asset_id = self.location.href.split('?')[0].split('/')[-1]

            # note: download data asset details
            asset_response = requests.get(
                self._wml_client.service_instance._href_definitions.get_data_asset_href(asset_id),
                params=self._wml_client._params(),
                headers=self._wml_client._get_headers())

            asset_details = self._wml_client.data_assets._handle_response(200, u'get assets', asset_response)

            attachment_id = asset_details['attachments'][0]['id']
            response = requests.get(
                self._wml_client.service_instance._href_definitions.get_attachment_href(asset_id, attachment_id),
                params=self._wml_client.data_assets._client._params(),
                headers=self._wml_client.data_assets._client._get_headers())

            if response.status_code == 200:
                try:
                    attachment_details = response.json()
                    if "url" in attachment_details:
                        file_asset_url = attachment_details['url']
                        if not file_asset_url.startswith("http"):
                            file_asset_url = self._wml_client.wml_credentials['url'] + file_asset_url
                        csv_response = requests.get(file_asset_url)

                        if csv_response.status_code != 200:
                            raise ApiRequestFailure(u'Failure during {}.'.format("downloading model"), csv_response)

                        downloaded_asset = csv_response.content

                        # note: read the csv/xlsx file from the memory directly into the pandas DataFrame
                        buffer = io.BytesIO(downloaded_asset)
                        data = try_load_dataset(buffer=buffer,
                                                sheet_name=self.auto_pipeline_params.get('excel_sheet', 0),
                                                separator=self.auto_pipeline_params.get('csv_separator', ','),
                                                encoding=self.auto_pipeline_params.get('encoding', 'utf-8')
                                                )

                        return data

                except requests.exceptions.MissingSchema:
                    pass  # go to 'handle' part and check if we are able to download data asset from WS

            # note: read the csv url
            if 'handle' in asset_details['attachments'][0]:
                attachment_url = asset_details['attachments'][0]['handle']['key']

                # note: make the whole url pointing out the csv
                artifact_content_url = (
                    f"{self._wml_client.service_instance._href_definitions.get_wsd_model_attachment_href()}"
                    f"{attachment_url}")

                # note: stream the whole CSV file
                csv_response = requests.get(artifact_content_url,
                                            params=self._wml_client._params(),
                                            headers=self._wml_client._get_headers(),
                                            stream=True)

                if csv_response.status_code != 200:
                    raise ApiRequestFailure(u'Failure during {}.'.format("downloading model"), csv_response)

                downloaded_asset = csv_response.content

                # note: read the csv/xlsx file from the memory directly into the pandas DataFrame
                buffer = io.BytesIO(downloaded_asset)
                data = try_load_dataset(buffer=buffer,
                                        sheet_name=self.auto_pipeline_params.get('excel_sheet', 0),
                                        separator=self.auto_pipeline_params.get('csv_separator', ','),
                                        encoding=self.auto_pipeline_params.get('encoding', 'utf-8')
                                        )

                return data

            else:
                # NFS scenario
                connection_id = asset_details['attachments'][0]['connection_id']
                connection_path = asset_details['attachments'][0]['connection_path']

                return self._download_data_from_nfs_connection_using_id_and_path(connection_id, connection_path)

        else:
            try:
                from project_lib import Project
                project = Project.access()

            except ModuleNotFoundError:
                raise NotImplementedError(f"This functionality can be run only on Watson Studio.")

            asset_id = self.location.href.split('?')[0].split('/')[-1]
            assets_list = project.get_assets()

            data_asset_name = None
            for asset in assets_list:
                if asset['asset_id'] == asset_id:
                    data_asset_name = asset['name']

            if data_asset_name is None:
                raise FileNotFoundError(f"Cannot find data asset with id: {asset_id}")

            buffer = project.get_file(data_asset_name)
            data = try_load_dataset(buffer=buffer,
                                    sheet_name=self.auto_pipeline_params.get('excel_sheet', 0),
                                    separator=self.auto_pipeline_params.get('csv_separator', ','),
                                    encoding=self.auto_pipeline_params.get('encoding', 'utf-8')
                                    )

            return data

    def _download_training_data_from_file_system(self, binary: bool = False) -> Tuple['DataFrame', bytes]:
        """Download training data for this connection. File system version."""

        try:
            url = self._wml_client.service_instance._href_definitions.get_wsd_model_attachment_href() + f"/{self.location.path.split('/assets/')[-1]}"
            # note: stream the whole CSV file
            csv_response = requests.get(url,
                                        params=self._wml_client._params(),
                                        headers=self._wml_client._get_headers(),
                                        stream=True)

            if csv_response.status_code != 200:
                raise ApiRequestFailure(u'Failure during {}.'.format("downloading model"), csv_response)

            downloaded_asset = csv_response.content

            if binary:
                return downloaded_asset

            # note: read the csv/xlsx file from the memory directly into the pandas DataFrame
            buffer = io.BytesIO(downloaded_asset)
            data = try_load_dataset(buffer=buffer,
                                    sheet_name=self.auto_pipeline_params.get('excel_sheet', 0),
                                    separator=self.auto_pipeline_params.get('csv_separator', ','),
                                    encoding=self.auto_pipeline_params.get('encoding', 'utf-8')
                                    )
        except (ApiRequestFailure, AttributeError):
            with open(self.location.path, 'rb') as data_buffer:
                if binary:
                    return data_buffer.read()

                data = try_load_dataset(buffer=data_buffer,
                                        sheet_name=self.auto_pipeline_params.get('excel_sheet', 0),
                                        separator=self.auto_pipeline_params.get('csv_separator', ','),
                                        encoding=self.auto_pipeline_params.get('encoding', 'utf-8')
                                        )

        return data

    def _download_indices_from_file_system(self, location_path: str) -> 'DataFrame':
        """Download indices for this connection. File system version."""

        try:
            url = self._wml_client.service_instance._href_definitions.get_wsd_model_attachment_href() + f"/{location_path.split('/assets/')[-1]}"
            # note: stream the whole CSV file
            csv_response = requests.get(url,
                                        params=self._wml_client._params(),
                                        headers=self._wml_client._get_headers(),
                                        stream=True)

            if csv_response.status_code != 200:
                raise ApiRequestFailure(u'Failure during {}.'.format("downloading model"), csv_response)

            downloaded_asset = csv_response.content
            # note: read the csv/xlsx file from the memory directly into the pandas DataFrame
            buffer = io.BytesIO(downloaded_asset)
            data = try_load_dataset(buffer=buffer,
                                    sheet_name=self.auto_pipeline_params.get('excel_sheet', 0),
                                    separator=self.auto_pipeline_params.get('csv_separator', ','),
                                    encoding=self.auto_pipeline_params.get('encoding', 'utf-8')
                                    )
        except (ApiRequestFailure, AttributeError):
            with open(location_path, 'rb') as data_buffer:
                data = try_load_tar_gz(buffer=data_buffer,
                                       separator=self.auto_pipeline_params.get('csv_separator', ','),
                                       encoding=self.auto_pipeline_params.get('encoding', 'utf-8')
                                       )

        return data

    def _download_obm_data_from_file_system(self) -> 'DataFrame':
        """Download preprocessed OBM data. FS version."""

        # note: fetch all OBM file part names
        url = self._wml_client.service_instance._href_definitions.get_wsd_model_attachment_href() + f"/auto_ml/{self.location.path.split('/auto_ml/')[-1]}/{self._run_id}/data/obm/features"
        params = self._wml_client._params()
        params['flat'] = "true"

        response = requests.get(url,
                                params=params,
                                headers=self._wml_client._get_headers())

        if response.status_code != 200:
            raise ApiRequestFailure(u'Failure during {}.'.format("getting files information"), response)

        file_names = [e['path'].split('/')[-1] for e in response.json()['resources'] if
                      e['type'] == 'file' and e['path'].split('/')[-1].startswith('part')]

        # TODO: this can be done simultaneously (multithreading / multiprocessing)
        # note: download all data parts and concatenate them into one output
        parts = []
        for file_name in file_names:
            csv_response = requests.get(url + '/' + file_name,
                                        params=self._wml_client._params(),
                                        headers=self._wml_client._get_headers(),
                                        stream=True)

            if csv_response.status_code != 200:
                raise ApiRequestFailure(u'Failure during {}.'.format("downloading model"), csv_response)

            downloaded_asset = csv_response.content
            # note: read the csv/xlsx file from the memory directly into the pandas DataFrame
            buffer = io.BytesIO(downloaded_asset)
            parts.append(try_load_dataset(buffer=buffer))

        data = concat(parts)
        # --- end note
        return data

    def _download_data_from_nfs_connection(self) -> 'DataFrame':
        """Download training data for this connection. NFS."""

        # note: as we need to load a data into the memory,
        # we are using pure requests and helpers from the WML client
        data_path = self.location.path
        connection_id = self.connection.asset_id

        return self._download_data_from_nfs_connection_using_id_and_path(connection_id, data_path)

    def _download_data_from_nfs_connection_using_id_and_path(self, connection_id, connection_path) -> 'DataFrame':
        """Download training data for this connection. NFS."""

        # it means that it is on ICP env and it is before fit, so let's throw error
        if not self._wml_client:
            raise CannotReadSavedRemoteDataBeforeFit()

        buffer = None
        # Note: workaround with volumes API as connections API changes data format
        try:
            connection_details = self._wml_client.connections.get_details(connection_id)
        except ApiRequestFailure as conn_error:
            if os.environ.get('TRAINING_NFS_PATH'):
                # Note: Only viable on AutoAI runtime
                base_path = os.environ.get('TRAINING_NFS_PATH') + "/0"
                if connection_path.startswith('/'):
                    data_path = f"{base_path}{connection_path}"
                else:
                    data_path = f"{base_path}/{connection_path}"
                with open(data_path, 'rb') as f:
                    buffer = io.BytesIO(f.read())
            else:
                raise conn_error

        if buffer is None:
            href = self._wml_client.volumes._client.service_instance._href_definitions.volume_upload_href(
                connection_details['entity']['properties']['volume'])
            full_href = f"{href[:-1]}{connection_path}"

            csv_response = requests.get(full_href,
                                        headers=self._wml_client._get_headers(),
                                        stream=True)

            # Note: if file is written in directory we need to create different href for download
            if csv_response.status_code != 200:
                path_parts = connection_path.split('/')
                full_href = f"{href[:-1]}{'/'.join(path_parts[:-1])}%2F{path_parts[-1]}"

                csv_response = requests.get(full_href,
                                            headers=self._wml_client._get_headers(),
                                            stream=True)

                if csv_response.status_code != 200:
                    raise ApiRequestFailure(u'Failure during {}.'.format("downloading data"), csv_response)

            downloaded_asset = csv_response.content
            # note: read the csv/xlsx file from the memory directly into the pandas DataFrame
            buffer = io.BytesIO(downloaded_asset)

        data = try_load_dataset(buffer=buffer,
                                sheet_name=self.auto_pipeline_params.get('excel_sheet', 0),
                                separator=self.auto_pipeline_params.get('csv_separator', ','),
                                encoding=self.auto_pipeline_params.get('encoding', 'utf-8')
                                )

        return data

    def _download_data_from_cos(self, cos_client: 'resource') -> 'DataFrame':
        """Download training data for this connection. COS version"""
        location = self.location.file_name if hasattr(self.location, "file_name") else self.location.path
        try:
            file = cos_client.Object(self.location.bucket,
                                     location).get()
        except Exception as e:
            cos_objects_list = list(cos_client.Bucket(self.location.bucket).objects.filter(
                Prefix=location))
            if len(cos_objects_list) > 0:
                file = list(cos_client.Bucket(self.location.bucket).objects.filter(
                    Prefix=location))[0].get()
            else:
                raise e

        buffer = io.BytesIO(file['Body'].read())
        data = try_load_dataset(buffer=buffer,
                                sheet_name=self.auto_pipeline_params.get('excel_sheet', 0),
                                separator=self.auto_pipeline_params.get('csv_separator', ','),
                                encoding=self.auto_pipeline_params.get('encoding', 'utf-8')
                                )

        return data

    def _create_conn_details_for_container(self) -> dict:
        if self._wml_client is not None:
            if self._wml_client.default_space_id is not None:
                details = self._wml_client.spaces.get_details(self._wml_client.default_space_id)

            else:
                project_url = self._wml_client.service_instance._href_definitions.get_project_href(
                    self._wml_client.default_project_id)

                response = requests.get(
                    project_url,
                    headers=self._wml_client._get_headers(projects_token=True)
                )

                details = response.json()

            properties = details['entity']['storage']['properties']
            creds = details['entity']['storage']['properties']['credentials'].get('admin')
            creds = creds if creds else details['entity']['storage']['properties']['credentials'].get('editor')
            properties.update(creds)
            properties['url'] = properties['endpoint_url']
            properties['access_key'] = properties.get('access_key_id')
            properties['secret_key'] = properties.get('secret_access_key')

            connection_details = {
                'entity': {
                    'datasource_type': 'bluemixcloudobjectstorage',
                    'properties': properties
                }
            }

        else:
            try:
                from project_lib import Project
                token = self._get_token_from_environment()

                if token is None:
                    raise NotImplementedError(
                        f"""To succesfully read the training data used in AutoAI experiment, you need to provide the project token.
                    **To insert the project token to your notebook:**
                        Click the More icon on your notebook toolbar and then click Insert project token.
                        Run the inserted code cell.
                    Note:
                    If you are told in a message that no project token exists, click the link in the message to be redirected to the project's Settings page where you can create a project token.
                    **To create a project token:**
                        Click New token in the Access tokens section on the Settings page of your project.
                        Enter a name, select Editor role for the project, and create a token.
                        Go back to your notebook, click the More icon on the notebook toolbar and then click Insert project token.
                        Run the inserted code cell."""
                    )

                token = token.split('Bearer ')[-1]

                _project = Project(project_id=os.environ.get('PROJECT_ID'),
                                   project_access_token=token)

                details = _project.get_storage_metadata()

                properties = details['properties']
                properties.update(properties['credentials']['editor'])
                properties['url'] = properties['endpoint_url']
                properties['access_key'] = properties.get('access_key_id')
                properties['secret_key'] = properties.get('secret_access_key')

                connection_details = {
                    'entity': {
                        'datasource_type': 'bluemixcloudobjectstorage',
                        'properties': properties
                    }
                }

            except ModuleNotFoundError:
                raise NotImplementedError(f"This functionality can be run only on Watson Studio.")

        return connection_details

    def _download_data_from_flight_service(self,
                                           data_location: 'DataConnection',
                                           binary: bool = False,
                                           read_to_file: str = None,
                                           flight_parameters: dict = None,
                                           headers: dict = None,
                                           number_of_batch_rows: int = None,
                                           sample_size_limit: int = DEFAULT_SAMPLE_SIZE_LIMIT,
                                           sampling_type: str = DEFAULT_SAMPLING_TYPE,
                                           return_data_as_iterator: bool = False,
                                           enable_sampling: bool = True,
                                           _return_subsampling_stats: bool = False):

        is_lib_installed(lib_name='pyarrow', minimum_version='3.0.0', install=True)

        # import the class only if flight scenario is enabled - do not import it in main import section
        from ibm_watson_machine_learning.data_loaders.experiment import ExperimentDataLoader
        from ibm_watson_machine_learning.data_loaders.datasets.experiment import ExperimentIterableDataset

        if flight_parameters is None:
            flight_parameters = {"num_partitions": 4}

        dict_connection = data_location._to_dict()

        experiment_metadata = {
            "n_parallel_data_connections": flight_parameters.get("num_partitions", 4),
            "prediction_column": self.auto_pipeline_params.get('prediction_column'),
            "prediction_type": self.auto_pipeline_params.get('prediction_type'),
            "project_id": self._wml_client.default_project_id,
            "space_id": self._wml_client.default_space_id,
            'headers': self._wml_client._get_headers() if headers is None else headers
        }

        experiment_metadata.update(self.auto_pipeline_params)

        experiment_iterable_dataset_setup_parameters = dict(
            connection=dict_connection,
            enable_sampling=enable_sampling if not binary else False,  # don't use sampling to read binary
            experiment_metadata=experiment_metadata,
            binary_data=binary,
            read_to_file=read_to_file,
            flight_parameters=flight_parameters if flight_parameters is not None else {},
            fallback_to_one_connection=False,
            _return_subsampling_stats=_return_subsampling_stats,
            number_of_batch_rows=number_of_batch_rows,
            sampling_type=sampling_type,
            sample_size_limit=sample_size_limit,
            _wml_client=self._wml_client
        )

        iterable_dataset = ExperimentIterableDataset(**experiment_iterable_dataset_setup_parameters)
        data_loader = ExperimentDataLoader(dataset=iterable_dataset)

        try:

            if not return_data_as_iterator:
                for data in data_loader:
                    return data
            else:
                return data_loader
        except TypeError as e1:
            # note: retry if there is problem with types:
            # try download data with  infer_as_varchar set to 'true' if some error occurs
            # final data downloaded and converted to proper types might has smaller size than sample_size_limit,
            # because data limit is calculated on data downloaded as varchar, before the conversion to more optimal type.
            try:
                iterable_dataset = ExperimentIterableDataset(**experiment_iterable_dataset_setup_parameters)

                iterable_dataset.connection.infer_as_varchar = 'true'
                data_loader = ExperimentDataLoader(dataset=iterable_dataset)

                if not return_data_as_iterator:
                    for data in data_loader:
                        return data
                else:
                    return data_loader
            except Exception as e2:
                raise DataStreamError(f"First attempt of downloading data failed with error: {e1}. \n"
                                      f"Retry and use infer as varchar also failed with error: {e2}.")

    def _upload_data_via_flight_service(self,
                                        data_location: 'DataConnection',
                                        data: DataFrame = None,
                                        file_path: str = None,
                                        remote_name: str = None,
                                        flight_parameters: dict = None,
                                        headers: dict = None):
        is_lib_installed(lib_name='pyarrow', minimum_version='3.0.0', install=True)

        # import the class only if flight scenario is enabled - do not import it in main import section
        from ibm_watson_machine_learning.data_loaders.datasets.experiment import ExperimentIterableDataset

        if flight_parameters is None:
            flight_parameters = {"num_partitions": 1}

        dict_connection = data_location._to_dict()

        if remote_name:
            dict_connection['location']['path'] = self._get_path_with_remote_name(dict_connection, remote_name)
        elif dict_connection.get('location', {}).get('file_name'):
            dict_connection['location']['path'] = dict_connection['location']['file_name']

        experiment_metadata = {
            "n_parallel_data_connections": flight_parameters.get("num_partitions", 1),
            "project_id": self._wml_client.default_project_id,
            "space_id": self._wml_client.default_space_id,
            'headers': self._wml_client._get_headers() if headers is None else headers
        }

        experiment_metadata.update(self.auto_pipeline_params)

        iterable_dataset = ExperimentIterableDataset(
            connection=dict_connection,
            experiment_metadata=experiment_metadata,
            binary_data=True if file_path is not None else False,
            flight_parameters=flight_parameters if flight_parameters is not None else {},
            _wml_client=self._wml_client
        )

        if data is not None:
            try:
                iterable_dataset.write(data=data)

            except Exception as e:
                if 'gRPC message exceeds maximum size' in str(e):
                    raise ValueError(f"Exceeds maximum data size. Please provide data file path "
                                     f"instead of the pandas DataFrame to upload data in binary mode. Error: {e}")

                else:
                    raise e

        else:
            iterable_dataset.write(file_path=file_path)

    @staticmethod
    def _get_path_with_remote_name(dict_connection: dict, remote_name: str) -> str:
        if dict_connection.get('location', {}).get('path'):
            updated_path = dict_connection['location']['path'] + '/' + remote_name
            updated_path = updated_path.replace('//', '/')
        elif dict_connection.get('location', {}).get('file_name'):
            actual_path = dict_connection['location']['file_name']
            last_slash_index = actual_path.rfind('/') if '/' in actual_path else 0

            if '.' in actual_path[last_slash_index:]:
                actual_path = actual_path[:-last_slash_index]

            if actual_path:
                updated_path = actual_path + '/' + remote_name
                updated_path = updated_path.replace('//', '/')
            else:
                updated_path = remote_name
        else:
            updated_path = remote_name

        return updated_path

    @staticmethod
    def _get_token_from_environment():
        if os.environ.get('RUNTIME_ENV_ACCESS_TOKEN_FILE'):
            with open(os.environ.get('RUNTIME_ENV_ACCESS_TOKEN_FILE'), 'r') as f:
                token = f.read()
        else:
            token = os.environ.get('USER_ACCESS_TOKEN')

        if token:
            token = token.replace('Bearer ', '')

        return token



#  (C) Copyright IBM Corp. 2021，2022.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging


logger = logging.getLogger(__name__)


class DataHandler:

    raise Exception("This data handler import path is deprecated. "
                    "Please do \"from ibm_watson_machine_learning.federated_learning.data_handler "
                    "import DataHandler\" instead ")

    def __init__(self, **kwargs):
        pass





#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import print_function
import ibm_watson_machine_learning._wrappers.requests as requests
import json
from ibm_watson_machine_learning.utils import SPACES_IMPORTS_DETAILS_TYPE, SPACES_EXPORTS_DETAILS_TYPE, SPACES_DETAILS_TYPE, INSTANCE_DETAILS_TYPE, MEMBER_DETAILS_TYPE, STR_TYPE, STR_TYPE_NAME, docstring_parameter, meta_props_str_conv, str_type_conv, get_file_from_cos, print_text_header_h2
from ibm_watson_machine_learning.metanames import SpacesMetaNames, MemberMetaNames, ExportMetaNames
from ibm_watson_machine_learning.href_definitions import is_uid
from ibm_watson_machine_learning.wml_resource import WMLResource
from ibm_watson_machine_learning.wml_client_error import  WMLClientError


_DEFAULT_LIST_LENGTH = 50


class Spaces(WMLResource):
    """
    Store and manage your spaces. This is applicable only for IBM Cloud Pak® for Data
    """
    ConfigurationMetaNames = SpacesMetaNames()
    MemberMetaNames = MemberMetaNames()
    ExportMetaNames = ExportMetaNames()
    """MetaNames for spaces creation."""

    def __init__(self, client):
        WMLResource.__init__(self, __name__, client)


        self._ICP = client.ICP

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def store(self, meta_props):
        """
                Create a space.

                **Parameters**

                .. important::
                   #. **meta_props**:  meta data of the space configuration. To see available meta names use:\n
                                    >>> client.spaces.ConfigurationMetaNames.get()

                      **type**: dict\n

                **Output**

                .. important::

                    **returns**: metadata of the stored space\n
                    **return type**: dict\n

                **Example**

                 >>> metadata = {
                 >>>  client.spaces.ConfigurationMetaNames.NAME: 'my_space',
                 >>>  client.spaces.ConfigurationMetaNames.DESCRIPTION: 'spaces',
                 >>> }
                 >>> spaces_details = client.spaces.store(meta_props=metadata)
                 >>> spaces_href = client.spaces.get_href(spaces_details)
        """

        # quick support for COS credentials instead of local path
        # TODO add error handling and cleaning (remove the file)
        WMLResource._chk_and_block_create_update_for_python36(self)
        Spaces._validate_type(meta_props, u'meta_props', dict, True)
        space_meta = self.ConfigurationMetaNames._generate_resource_metadata(
            meta_props,
            with_validation=True,
            client=self._client

        )

        creation_response = requests.post(
            self._wml_credentials['url'] + '/v4/spaces',
            headers=self._client._get_headers(),
            json=space_meta
        )

        spaces_details = self._handle_response(201, u'creating new spaces', creation_response)

        # Cloud Convergence: Set self._client.wml_credentials['instance_id'] to instance_id
        # during client.set.default_space since that's where space is associated with client
        # and also in client.set.default_project

        return spaces_details

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_href(spaces_details):
        """
            Get space_href from space details.

            **Parameters**

            .. important::
                #. **space_details**:  Metadata of the stored space\n
                   **type**: dict\n

            **Output**

            .. important::
                **returns**: space href\n
                **return type**: str

            **Example**

             >>> space_details = client.spaces.get_details(space_uid)
             >>> space_href = client.spaces.get_href(deployment)
        """

        Spaces._validate_type(spaces_details, u'spaces_details', object, True)
        Spaces._validate_type_of_details(spaces_details, SPACES_DETAILS_TYPE)

        return WMLResource._get_required_element_from_dict(spaces_details, u'spaces_details',
                                                           [u'metadata', u'href'])

    @staticmethod
    def get_uid(spaces_details):
        """
            Get space_uid from space details.

            **Parameters**

            .. important::
                #. **space_details**:  Metadata of the stored space\n
                   **type**: dict\n

            **Output**

            .. important::
                **returns**: space UID\n
                **return type**: str

            **Example**

             >>> space_details = client.spaces.get_details(space_uid)
             >>> space_uid = client.spaces.get_uid(deployment)
        """

        Spaces._validate_type(spaces_details, u'spaces_details', object, True)
        Spaces._validate_type_of_details(spaces_details, SPACES_DETAILS_TYPE)

        return WMLResource._get_required_element_from_dict(spaces_details, u'spaces_details',
                                                           [u'metadata', u'guid'])

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def delete(self, space_uid):
        """
            Delete a stored space.

            **Parameters**

            .. important::
                #. **space_uid**:  space UID\n
                   **type**: str\n

            **Output**

            .. important::
                **returns**: status ("SUCCESS" or "FAILED")\n
                **return type**: str\n

            **Example**

             >>> client.spaces.delete(deployment_uid)
        """

        space_uid = str_type_conv(space_uid)
        Spaces._validate_type(space_uid, u'space_uid', STR_TYPE, True)

        space_endpoint = self._client.service_instance._href_definitions.get_space_href(space_uid)

        response_delete = requests.delete(space_endpoint, headers=self._client._get_headers())

        return self._handle_response(204, u'space deletion', response_delete, False)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_details(self, space_uid=None, limit=None):
        """
           Get metadata of stored space(s). If space UID is not specified, it returns all the spaces metadata.

           **Parameters**

           .. important::
                #. **space_uid**: Space UID (optional)\n
                   **type**: str\n
                #. **limit**:  limit number of fetched records (optional)\n
                   **type**: int\n

           **Output**

           .. important::
                **returns**: metadata of stored space(s)\n
                **return type**: dict
                dict (if UID is not None) or {"resources": [dict]} (if UID is None)\n

           .. note::
                If UID is not specified, all spaces metadata is fetched\n

           **Example**

            >>> space_details = client.spaces.get_details(space_uid)
            >>> space_details = client.spaces.get_details()
        """

        space_uid = str_type_conv(space_uid)
        Spaces._validate_type(space_uid, u'space_uid', STR_TYPE, False)
        Spaces._validate_type(limit, u'limit', int, False)

        href = self._client.service_instance._href_definitions.get_spaces_href()
        if space_uid is None:
            return self._get_no_space_artifact_details(href+"?include=name,tags,custom,description", None, limit, 'spaces')
        return self._get_no_space_artifact_details(href, space_uid, limit, 'spaces')

    def list(self, limit=None):
        """
           List stored spaces. If limit is set to None there will be only first 50 records shown.

           **Parameters**

           .. important::
                #. **limit**:  limit number of fetched records\n
                   **type**: int\n

           **Output**

           .. important::
                This method only prints the list of all spaces in a table format.\n
                **return type**: None\n

           **Example**

            >>> client.spaces.list()
        """

        space_resources = self.get_details(limit=limit)[u'resources']
        space_values = [(m[u'metadata'][u'guid'], m[u'entity'][u'name'], m[u'metadata'][u'created_at']) for m in space_resources]

        self._list(space_values, [u'GUID', u'NAME', u'CREATED'], limit, _DEFAULT_LIST_LENGTH)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def update(self, space_uid, changes):
        """
                Updates existing space metadata.

                **Parameters**

                .. important::
                    #. **space_uid**:  UID of space which definition should be updated\n
                       **type**: str\n
                    #. **changes**:  elements which should be changed, where keys are ConfigurationMetaNames\n
                       **type**: dict\n

                **Output**

                .. important::
                    **returns**: metadata of updated space\n
                    **return type**: dict\n

                **Example**

                 >>> metadata = {
                 >>> client.spaces.ConfigurationMetaNames.NAME:"updated_space"
                 >>> }
                 >>> space_details = client.spaces.update(space_uid, changes=metadata)
        """

        space_uid = str_type_conv(space_uid)
        self._validate_type(space_uid, u'space_uid', STR_TYPE, True)
        self._validate_type(changes, u'changes', dict, True)
        meta_props_str_conv(changes)

        details = self.get_details(space_uid)

        patch_payload = self.ConfigurationMetaNames._generate_patch_payload(details['entity'], changes,
                                                                            with_validation=True)

        href = self._client.service_instance._href_definitions.get_space_href(space_uid)

        response = requests.patch(href, json=patch_payload, headers=self._client._get_headers())

        updated_details = self._handle_response(200, u'spaces patch', response)

        return updated_details


#######SUPPORT FOR SPACE MEMBERS

    ###GET MEMBERS DETAILS
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_members_details(self, space_uid, member_id=None, limit=None):
        """
           Get metadata of members associated with a space. If member UID is not specified, it returns all the members metadata.

           **Parameters**

           .. important::
                #. **space_uid**: member UID (optional)\n
                   **type**: str\n
                #. **limit**:  limit number of fetched records (optional)\n
                   **type**: int\n

           **Output**

           .. important::
                **returns**: metadata of member(s) of a space\n
                **return type**: dict
                dict (if UID is not None) or {"resources": [dict]} (if UID is None)\n

           .. note::
                If member id is not specified, all members metadata is fetched\n

           **Example**

            >>> member_details = client.spaces.get_member_details(space_uid,member_id)
        """

        space_uid = str_type_conv(space_uid)
        Spaces._validate_type(space_uid, u'space_uid', STR_TYPE, True)

        member_uid = str_type_conv(member_id)
        Spaces._validate_type(member_id, u'member_id', STR_TYPE, False)

        Spaces._validate_type(limit, u'limit', int, False)

        href = self._client.service_instance._href_definitions.get_members_href(space_uid)

        return self._get_no_space_artifact_details(href, member_uid, limit, 'space members')

    ##DELETE MEMBERS

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def delete_members(self, space_uid,member_id):
        """
            Delete a member associated with a space.

            **Parameters**

            .. important::
                #. **space_uid**:  space UID\n
                   **type**: str\n
                #. **member_uid**:  member UID\n
                   **type**: str\n

            **Output**

            .. important::
                **returns**: status ("SUCCESS" or "FAILED")\n
                **return type**: str\n

            **Example**

             >>> client.spaces.delete_member(space_uid,member_id)
        """

        space_uid = str_type_conv(space_uid)
        Spaces._validate_type(space_uid, u'space_uid', STR_TYPE, True)

        member_id = str_type_conv(member_id)
        Spaces._validate_type(member_id, u'member_id', STR_TYPE, True)

        member_endpoint = self._client.service_instance._href_definitions.get_member_href(space_uid,member_id)

        response_delete = requests.delete(member_endpoint, headers=self._client._get_headers())

        return self._handle_response(204, u'space member deletion', response_delete, False)

#######UPDATE MEMBERS

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def update_member(self, space_uid, member_id, changes):
            """
                    Updates existing member metadata.

                    **Parameters**

                    .. important::
                        #. **space_uid**:  UID of space\n
                           **type**: str\n
                        #. **member_id**:  UID of member that needs to be updated\n
                           **type**: str\n
                        #. **changes**:  elements which should be changed, where keys are ConfigurationMetaNames\n
                           **type**: dict\n

                    **Output**

                    .. important::
                        **returns**: metadata of updated member\n
                        **return type**: dict\n

                    **Example**

                     >>> metadata = {
                     >>> client.spaces.ConfigurationMetaNames.ROLE:"viewer"
                     >>> }
                     >>> member_details = client.spaces.update_member(space_uid, member_id, changes=metadata)
            """

            space_uid = str_type_conv(space_uid)
            self._validate_type(space_uid, u'space_uid', STR_TYPE, True)
            member_id = str_type_conv(member_id)
            self._validate_type(member_id, u'member_id', STR_TYPE, True)

            self._validate_type(changes, u'changes', dict, True)
            meta_props_str_conv(changes)

            details = self.get_members_details(space_uid,member_id)

            patch_payload = self.MemberMetaNames._generate_patch_payload(details['entity'], changes,
                                                                                with_validation=True)

            href = self._client.service_instance._href_definitions.get_member_href(space_uid, member_id)

            response = requests.patch(href, json=patch_payload, headers=self._client._get_headers())

            updated_details = self._handle_response(200, u'members patch', response)

            return updated_details

#####CREATE MEMBER
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def create_member(self, space_uid,meta_props):
        """
                Create a member within a space.

                **Parameters**

                .. important::
                   #. **meta_props**:  meta data of the member configuration. To see available meta names use:\n
                                    >>> client.spaces.MemberMetaNames.get()

                      **type**: dict\n

                **Output**

                .. important::

                    **returns**: metadata of the stored member\n
                    **return type**: dict\n

                .. note::
                    * client.spaces.MemberMetaNames.ROLE can be any one of the following "viewer, editor, admin"\n
                    * client.spaces.MemberMetaNames.IDENTITY_TYPE can be any one of the following "user,service"\n
                    * client.spaces.MemberMetaNames.IDENTITY can be either service-ID or IAM-userID\n

                **Example**

                 >>> metadata = {
                 >>>  client.spaces.MemberMetaNames.ROLE:"Admin",
                 >>>  client.spaces.MemberMetaNames.IDENTITY:"iam-ServiceId-5a216e59-6592-43b9-8669-625d341aca71",
                 >>>  client.spaces.MemberMetaNames.IDENTITY_TYPE:"service"
                 >>> }
                 >>> members_details = client.spaces.create_member(space_uid=space_id, meta_props=metadata)
        """

        # quick support for COS credentials instead of local path
        # TODO add error handling and cleaning (remove the file)
        Spaces._validate_type(meta_props, u'meta_props', dict, True)
        space_meta = self.MemberMetaNames._generate_resource_metadata(
            meta_props,
            with_validation=True,
            client=self._client

        )

        creation_response = requests.post(
            self._wml_credentials['url'] + '/v4/spaces/'+space_uid+"/members",
            headers=self._client._get_headers(),
            json=space_meta
        )

        members_details = self._handle_response(201, u'creating new members', creation_response)

        return members_details

    def list_members(self, space_uid ,limit=None):
            """
               List stored members of a space. If limit is set to None there will be only first 50 records shown.

               **Parameters**

               .. important::
                    #. **limit**:  limit number of fetched records\n
                       **type**: int\n

               **Output**

               .. important::
                    This method only prints the list of all members associated with a space in a table format.\n
                    **return type**: None\n

               **Example**

                >>> client.spaces.list_members()
            """

            member_resources = self.get_members_details(space_uid,limit=limit)[u'resources']
            space_values = [(m[u'metadata'][u'guid'],  m[u'entity'][u'identity'], m[u'entity'][u'identity_type'], m[u'entity'][u'role'], m[u'metadata'][u'created_at']) for m in member_resources]

            self._list(space_values, [u'GUID', u'USERNAME', u'IDENTITY_TYPE', u'ROLE', u'CREATED'], limit, _DEFAULT_LIST_LENGTH)




    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_member_href(member_details):
        """
            Get member_href from member details.

            **Parameters**

            .. important::
                #. **space_details**:  Metadata of the stored member\n
                   **type**: dict\n

            **Output**

            .. important::
                **returns**: member href\n
                **return type**: str

            **Example**

             >>> member_details = client.spaces.get_member_details(member_id)
             >>> member_href = client.spaces.get_member_href(member_details)
        """

        Spaces._validate_type(member_details, u'member details', object, True)
        Spaces._validate_type_of_details(member_details, MEMBER_DETAILS_TYPE)

        return WMLResource._get_required_element_from_dict(member_details, u'member_details',
                                                           [u'metadata', u'href'])

    @staticmethod
    def get_member_uid(member_details):
        """
            Get member_uid from member details.

            **Parameters**

            .. important::
                #. **member_details**:  Metadata of the created member\n
                   **type**: dict\n

            **Output**

            .. important::
                **returns**: member UID\n
                **return type**: str

            **Example**

             >>> member_details = client.spaces.get_member_details(member_id)
             >>> member_id = client.spaces.get_member_uid(member_details)
        """

        Spaces._validate_type(member_details, u'member_details', object, True)
        Spaces._validate_type_of_details(member_details, MEMBER_DETAILS_TYPE)

        return WMLResource._get_required_element_from_dict(member_details, u'member_details',
                                                           [u'metadata', u'guid'])


    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def imports(self, space_uid, file_path):
        """
                Updates existing space metadata.
                Imports assets in the zip file to a space.

                **Parameters**

                .. important::
                    #. **space_uid**:  UID of space which definition should be updated\n
                       **type**: str\n
                    #. **file_path**:  Path to the content file to be imported\\n
                       **type**: dict\n

                **Output**

                .. important::
                    **returns**: metadata of import space\n
                    **return type**: str\n

                **Example**

                 >>> space_details = client.spaces.imports(space_uid, file_path="/tmp/spaces.zip")
        """

        space_uid = str_type_conv(space_uid)
        self._validate_type(space_uid, u'space_uid', STR_TYPE, True)
        self._validate_type(file_path, u'file_path', STR_TYPE, True)



        with open(str(file_path), 'rb') as archive:
            data = archive.read()

        href = self._client.service_instance._href_definitions.get_space_href(space_uid) + "/imports"

        response = requests.post(href, headers=self._client._get_headers(), data=data)

        import_space_details = self._handle_response(202, u'spaces import', response)

        return import_space_details



    @staticmethod
    def get_imports_uid(imports_space_details):
        """
            Get imports_uid from imports space details.

            **Parameters**

            .. important::
                #. **imports_space_details**:  Metadata of the created space import\n
                   **type**: dict\n

            **Output**

            .. important::
                **returns**: imports space UID\n
                **return type**: str

            **Example**

             >>> member_details = client.spaces.get_imports_details(space_uid, imports_id)
             >>> imports_id = client.spaces.get_imports_uid(imports_space_details)
        """

        Spaces._validate_type(imports_space_details, u'member_details', object, True)
        Spaces._validate_type_of_details(imports_space_details, SPACES_IMPORTS_DETAILS_TYPE)

        return WMLResource._get_required_element_from_dict(imports_space_details, u'imports_space_details',[u'metadata', u'guid'])

    @staticmethod
    def get_exports_uid(exports_space_details):
        """
            Get imports_uid from imports space details.

            **Parameters**

            .. important::
                #. **space_exports_details**:  Metadata of the created space import\n
                   **type**: dict\n

            **Output**

            .. important::
                **returns**: exports space UID\n
                **return type**: str

            **Example**

             >>> member_details = client.spaces.get_exports_details(space_uid, exports_id)
             >>> imports_id = client.spaces.get_imports_uid(exports_space_details)
        """

        Spaces._validate_type(exports_space_details, u'exports_space_details', object, True)
        Spaces._validate_type_of_details(exports_space_details, SPACES_EXPORTS_DETAILS_TYPE)

        return WMLResource._get_required_element_from_dict(exports_space_details, u'exports_space_details',
                                                           [u'metadata', u'guid'])

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_imports_details(self, space_uid, imports_id=None, limit=None):
        """
           Get metadata of stored space(s). If space UID is not specified, it returns all the spaces metadata.

           **Parameters**

           .. important::
                #. **space_uid**: Space UID (optional)\n
                   **type**: str\n
                #. **limit**:  limit number of fetched records (optional)\n
                   **type**: int\n

           **Output**

           .. important::
                **returns**: metadata of stored space(s)\n
                **return type**: dict
                dict (if UID is not None) or {"resources": [dict]} (if UID is None)\n

           .. note::
                If UID is not specified, all spaces metadata is fetched\n

           **Example**

            >>> space_details = client.spaces.get_imports_details(space_uid)
            >>> space_details = client.spaces.get_imports_details(space_uid,imports_id)
        """

        space_uid = str_type_conv(space_uid)
        Spaces._validate_type(space_uid, u'space_uid', STR_TYPE, False)
        Spaces._validate_type(imports_id, u'imports_uid', STR_TYPE, False)
        Spaces._validate_type(limit, u'limit', int, False)


        if imports_id is None:
            href = self._client.service_instance._href_definitions.get_space_href(space_uid) + "/imports"
        else:
            href =  self._client.service_instance._href_definitions.get_space_href(space_uid) + "/imports/" + imports_id

        response = requests.get(href, headers=self._client._get_headers())

        import_space_details = self._handle_response(200, u'spaces import details', response)

        return import_space_details

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def exports(self, space_uid, meta_props):
        """
                Updates existing space metadata.
                Exports assets in the zip file from a space.

                **Parameters**

                .. important::
                   #. **meta_props**:  meta data of the space configuration. To see available meta names use:\n
                                    >>> client.spaces.ExportMetaNames.get()

                      **type**: dict\n

                **Output**

                .. important::
                    **returns**: metadata of import space\n
                    **return type**: str\n

                **Example**
                    >>> meta_props = {
                    >>>     client.spaces.ExportMetaNames.NAME: "sample",
                    >>>     client.spaces.ExportMetaNames.DESCRIPTION : "test description",
                    >>>     client.spaces.ExportMetaNames.ASSETS : {"data_assets": [], "wml_model":[]} }
                    >>> }
                    >>> space_details = client.spaces.exports(space_uid, meta_props=meta_props)

        """

        Spaces._validate_type(meta_props, u'meta_props', dict, True)
        space_exports_meta = self.ExportMetaNames._generate_resource_metadata(
            meta_props,
            with_validation=True,
            client=self._client)

        space_exports_meta_json = json.dumps(space_exports_meta)

        space_uid = str_type_conv(space_uid)
        self._validate_type(space_uid, u'space_uid', STR_TYPE, True)

        href = self._client.service_instance._href_definitions.get_space_href(space_uid) + "/exports"

        response = requests.post(href, headers=self._client._get_headers(), data=space_exports_meta_json)

        export_space_details = self._handle_response(202, u'spaces export', response)

        return export_space_details


    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_exports_details(self, space_uid, exports_id=None, limit=None):
        """
           Get details of exports for space. If exports UID is not specified, it returns all the spaces metadata.

           **Parameters**

           .. important::
                #. **space_uid**: Space UID (optional)\n
                   **type**: str\n
                #. **limit**:  limit number of fetched records (optional)\n
                   **type**: int\n
           **Output**

           .. important::
                **returns**: metadata of stored space(s)\n
                **return type**: dict
                dict (if UID is not None) or {"resources": [dict]} (if UID is None)\n

           .. note::
                If UID is not specified, all spaces metadata is fetched\n

           **Example**

            >>> space_details = client.spaces.get_exports_details(space_uid)
            >>> space_details = client.spaces.get_exports_details(space_uid,exports_id)

        """

        space_uid = str_type_conv(space_uid)
        Spaces._validate_type(space_uid, u'space_uid', STR_TYPE, False)
        Spaces._validate_type(exports_id, u'imports_uid', STR_TYPE, False)
        Spaces._validate_type(limit, u'limit', int, False)


        if exports_id is None:
            href = self._client.service_instance._href_definitions.get_space_href(space_uid) + "/exports"
        else:
            href =  self._client.service_instance._href_definitions.get_space_href(space_uid) + "/exports/" + exports_id

        response = requests.get(href, headers=self._client._get_headers())

        export_space_details = self._handle_response(200, u'spaces exports details', response)

        return export_space_details


    def download(self, space_uid, space_exports_uid, filename=None):
        """
        Downloads zip file deployment of specified UID.

        **Parameters**

        .. important::

            #. **exports_space_uid**:  UID of virtual deployment.\n
               **type**: str\n

            #. **filename**: Filename of downloaded archive. (optional)\n
               **type**: str\n

        **Output**

        .. important::

            **returns**: Path to downloaded file.\n
            **return type**: str\n

        **Example**

         >>> client.spaces.download(space_uid)
        """

        space_exports_uid = str_type_conv(space_exports_uid)
        Spaces._validate_type(space_exports_uid, u'space_exports_uid', STR_TYPE, False)

        if space_exports_uid is not None and not is_uid(space_exports_uid):
            raise WMLClientError(u'\'space_exports_uid\' is not an uid: \'{}\''.format(space_exports_uid))

        href =  self._client.service_instance._href_definitions.get_space_href(space_uid) + "/exports/" + space_exports_uid + "/content"

        response = requests.get(
            href,
            headers=self._client._get_headers()
        )

        if filename is None:
            filename = 'wmlspace.zip'

        if response.status_code == 200:
            with open(filename, "wb") as new_file:
                new_file.write(response.content)
                new_file.close()

                print_text_header_h2(
                    u'Successfully downloaded spaces export file: ' + str(filename))

                return filename
        else:
            raise WMLClientError(u'Unable to download spaces export: ' + response.text)




#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ibm_watson_machine_learning.libs.repo.mlrepository import MetaNames
from ibm_watson_machine_learning.libs.repo.mlrepository import MetaProps
from ibm_watson_machine_learning.libs.repo.mlrepository import ModelArtifact
from ibm_watson_machine_learning.libs.repo.mlrepositoryartifact.tensorflow_pipeline_reader import TensorflowPipelineReader
from ibm_watson_machine_learning.libs.repo.mlrepositoryartifact.generic_file_reader import GenericFileReader
from ibm_watson_machine_learning.libs.repo.mlrepositoryartifact.version_helper import TensorflowVersionHelper
from ibm_watson_machine_learning.libs.repo.util.library_imports import LibraryChecker
from ibm_watson_machine_learning.libs.repo.base_constants import *
from .python_version import PythonVersion
from ibm_watson_machine_learning.libs.repo.util.unique_id_gen import uid_generate
from ibm_watson_machine_learning.libs.repo.util.exceptions import UnmatchedKerasVersion

import os, shutil, tarfile, json

lib_checker = LibraryChecker()

if lib_checker.installed_libs[TENSORFLOW]:
    import tensorflow as tf

class TensorflowPipelineModelArtifact(ModelArtifact):
    """
    Class of Tensorflow model artifacts created with MLRepositoryCLient.

    """
    def __init__(self,tensorflow_pipeline_model,
                 signature_def_map,
                 tags=None,
                 assets_collection=None,
                 legacy_init_op=None,
                 clear_devices=False,
                 main_op=None,
                 uid=None, name=None, meta_props=MetaProps({}),):
        lib_checker.check_lib(TENSORFLOW)
        super(TensorflowPipelineModelArtifact, self).__init__(uid, name, meta_props)

        if '2.1.0' in tf.__version__ and not isinstance(tensorflow_pipeline_model,tf.compat.v1.Session):
            raise TypeError("sess should be of type : %s" % tf.compat.v1.Session)
        elif not isinstance(tensorflow_pipeline_model,tf.Session):
            raise TypeError("sess should be of type : %s" % tf.Session)

        self.ml_pipeline_model = tensorflow_pipeline_model
        self.signature_def_map = signature_def_map
        self.tags = tags
        self.assets_collection = assets_collection
        self.legacy_init_op = legacy_init_op
        self.clear_devices = clear_devices
        self.main_op = main_op

        self.ml_pipeline = None     # no pipeline or parent reference

        if meta_props.prop(MetaNames.RUNTIMES) is None and meta_props.prop(MetaNames.RUNTIME_UID) is None and meta_props.prop(MetaNames.FRAMEWORK_RUNTIMES) is None:
            ver = PythonVersion.significant()
            runtimes = '[{"name":"python","version": "'+ ver + '"}]'
            self.meta.merge(
                MetaProps({MetaNames.FRAMEWORK_RUNTIMES: runtimes})
            )

        self.meta.merge(
            MetaProps({
                MetaNames.FRAMEWORK_NAME:   TensorflowVersionHelper.model_type(tensorflow_pipeline_model),
                MetaNames.FRAMEWORK_VERSION: TensorflowVersionHelper.model_version(tensorflow_pipeline_model)
            })
        )

    def pipeline_artifact(self):
        """
        Returns None. Pipeline is not implemented for Tensorflow model.
        """
        pass

    def reader(self):
        """
        Returns reader used for getting pipeline model content.

        :return: reader for TensorflowPipelineModelArtifact.pipeline.Pipeline
        :rtype: TensorflowPipelineReader
        """
        try:
            return self._reader
        except:
            self._reader = TensorflowPipelineReader(self.ml_pipeline_model,
                                                    self.signature_def_map,
                                                    self.tags,
                                                    self.assets_collection,
                                                    self.legacy_init_op,
                                                    self.clear_devices,
                                                    self.main_op)
            return self._reader

    def _copy(self, uid=None, meta_props=None):
        if uid is None:
            uid = self.uid

        if meta_props is None:
            meta_props = self.meta

        return TensorflowPipelineModelArtifact(
            self.ml_pipeline_model,
            self.signature_def_map,
            self.tags,
            self.assets_collection,
            self.legacy_init_op,
            self.clear_devices,
            self.main_op,
            uid=uid,
            name=self.name,
            meta_props=meta_props
        )


class TensorflowPipelineModelTarArtifact(ModelArtifact):
    """
    Class of serialized Tensorflow model artifacts in tar.gz format and created
    with MLRepositoryCLient.

    """
    def __init__(self,
                 tensorflow_tar_artifact,
                 uid=None,
                 name=None,
                 meta_props=MetaProps({}),):
        if not (lib_checker.installed_libs[TENSORFLOW]):
            raise NameError("Please install Tensorflow package and re-execute the command")
        super(TensorflowPipelineModelTarArtifact, self).__init__(uid, name, meta_props)

        self.ml_pipeline_model = tensorflow_tar_artifact
        self.ml_pipeline = None     # no pipeline or parent reference

    def pipeline_artifact(self):
        """
        Returns None. Pipeline is not implemented for Tensorflow model.
        """
        pass

    def reader(self):
        """
        Returns reader used for getting pipeline model content.

        :return: reader for TensorflowPipelineModelArtifact.pipeline.Pipeline
        :rtype: TensorflowPipelineReader
        """
        try:
            return self._reader
        except:
            self._reader = GenericFileReader(self.ml_pipeline_model)
            return self._reader

    def _copy(self, uid=None, meta_props=None):
        if uid is None:
            uid = self.uid

        if meta_props is None:
            meta_props = self.meta

        return TensorflowPipelineModelTarArtifact(
            self.ml_pipeline_model,
            uid=uid,
            name=self.name,
            meta_props=meta_props
        )

    def unzip_artifact(self):
        try:
            tar = tarfile.open(self.ml_pipeline_model, 'r:gz')
        except Exception as ex:
            raise IOError('Unable to read the compressed archive file in {0} due to '
                          'error "{1}". '
                          'Ensure a valid tar archive is compressed in gzip format.'
                          .format(self.ml_pipeline_model, ex))
        return tar

    def get_keras_version(self):
        def extract_keras_version(extract_dir, h5_file_tar_info_name):
            import h5py
            keras_version_key = 'keras_version'
            file_name = os.path.join(extract_dir, h5_file_tar_info_name)
            try:
                with h5py.File(file_name) as f:
                    keras_version = f.attrs[keras_version_key] if keras_version_key in f.attrs else None
                    f.close()
            except Exception as ex:
                keras_version = None
                shutil.rmtree(extract_dir)
                raise Exception("Error while opening file to get Keras version. Error message - '{}' ".format(str(ex)))
            return keras_version.decode('utf-8')

        extract_dir = uid_generate(5)
        tar = self.unzip_artifact()

        # Return None if the artifact is not a valid keras archive
        if len(tar.getmembers()) != 1:
            return None

        h5_file_tar_info = tar.getmembers()[0]
        if os.path.exists(extract_dir):
            shutil.rmtree(extract_dir)
        os.makedirs(os.path.join("./", extract_dir))
        tar.extract(h5_file_tar_info, extract_dir)
        keras_version = extract_keras_version(extract_dir, h5_file_tar_info.name)
        shutil.rmtree(extract_dir)
        return keras_version

    def is_valid_tf_archive(self):
        expected_file_name = "saved_model.pb"
        expected_file_with_path = os.path.join('.', expected_file_name)
        tar = self.unzip_artifact()
        is_valid_tf = True if expected_file_with_path in [file.name for file in tar.getmembers()] else False
        tar.close()
        return is_valid_tf

    def update_keras_version_meta(self, keras_version):
        KERAS_FRAMEWORK_NAME = "keras"
        # check if meta prop has already an entry for keras
        framewk_library_entry = {"name": KERAS_FRAMEWORK_NAME,
                                  "version": keras_version}
        framewk_library_entry_full = json.dumps([framewk_library_entry])

        if MetaNames.FRAMEWORK_LIBRARIES not in self.meta.meta:
            self.meta.add(MetaNames.FRAMEWORK_LIBRARIES, framewk_library_entry_full)
        else:
            lib_entry = json.loads(self.meta.meta[MetaNames.FRAMEWORK_LIBRARIES])
            for lib in lib_entry:
                if lib['name'] == KERAS_FRAMEWORK_NAME:
                    if lib['version'] != keras_version:
                        raise UnmatchedKerasVersion('Keras version specified as metadata and Keras version used to '
                                                    'train the model does not match. Keras version specified in '
                                                    'metadata is {0} . Keras version used to train the model is '
                                                    '{1}'.format(lib['version'], keras_version))
                    else:
                        return self
            lib_entry.append(framewk_library_entry)
            self.meta.meta.update({MetaNames.FRAMEWORK_LIBRARIES: json.dumps(lib_entry)})
        return self



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ibm_watson_machine_learning.libs.repo.mlrepository import ArtifactReader
import time

class ContentReader(ArtifactReader):
    """
    Reader which read content of pipeline/pipeline model from repository using href.

    :param str content_href: href to content, returned by repository api
    :param MLRepositoryApi: repository api object
    """
    def __init__(self, content_href, repository_api, download_href=None):
        self.content_href = content_href
        self.download_href = download_href
        self.repository_api = repository_api
        # val apiInvoker: ApiInvoker = repositoryApi.apiInvoker
        # val basePath: String = repositoryApi.basePath
        # var entity: Option[InputStream] = None
        self.connection = None

    def read(self):
        """
        Returns stream object with content of pipeline/pipeline model.

        :return: binary stream
        :rtype: HTTPResponse (from urllib3)
        """
        if self.connection is not None:
            self.close()
        self.connection = self._download()
        return self.connection

    def close(self):  # TODO ????
        """
        Closes stream to content.
        """
        # entity.map(is => is.close())
        self.connection.release_conn()

    def _download(self):
        model_size_1GB = 1073741824
        model_content_size = 0

        if (self.content_href.__contains__("models")):
            if (self.download_href is not None):
                return self.repository_api.download_artifact_content_v4_cloud(self.download_href, 'true')
            else:
                return self.repository_api.download_artifact_content(self.content_href, 'false')

        if self.content_href.__contains__("libraries"):
            return self.repository_api.download_artifact_content(self.content_href, 'false', accept='application/gzip')

        if self.content_href.__contains__("runtimes"):
            return self.repository_api.download_artifact_content(self.content_href, 'false', accept='text/plain')
        else:
            #downloading training definition content or modelsize is less than model_size_1GB
            return self.repository_api.download_artifact_content(self.content_href, 'false')



#  (C) Copyright IBM Corp. 2021，2022.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import abc
import numpy

from ibm_watson_machine_learning.federated_learning.data_util import get_min, get_max, get_mean,\
    get_var, get_std, get_quantile, get_normalizer, get_standardscaler, \
    get_minmaxscaler
from ibm_watson_machine_learning.federated_learning.FLExceptions import FLException

logger = logging.getLogger(__name__)


class DataHandler(abc.ABC):
    """
    Base class to load and pre-process data.
    """

    def __init__(self, **kwargs):
        self.x_train = None
        self.y_train = None
        self.x_test = None
        self.y_test = None
        self.preprocessor = None
        self.testing_generator = None
        self.training_generator = None

    @abc.abstractmethod
    def get_data(self, **kwargs):
        """
        Access the local dataset and return the training and testing dataset
        as a tuple.

        :param kwargs:
        :return: `tuple`. (training_set, testing_set)
        """
        raise NotImplemented

    def get_val_data(self, **kwargs):
        """
        Access the local dataset and return the validation dataset
        as a tuple.
        :param kwargs:
        :return: `tuple`. (validation_set)
        """
        pass

    def get_statistics_of_training_data(self, sample_data_schema,
                                        lst_stats_name, **kwargs):
        """
        Return the corresponding statistics, which is specified by the
        provided list of statistics names, of the local training dataset.

        :param sample_data_schema: Provided data with only feature values. \
        Assuming the dataset has shape (num_samples, num_features).
        :type sample_data_schema: `np.array`
        :param lst_stats_name: A list of statistics names, \
        all in lowercase form, for example, ['min'], ['mean', 'variance'], etc.
        :type lst_stats_name: `list` of `str`
        :param kwargs: Additional parameters to obtain the statistics.
        :type kwargs: `dict`
        :return: The requested statistics based on the local dataset.
        :rtype: `dict`
        """
        if sample_data_schema is None and self.x_train is None:
            raise FLException("No data is provided!")
        elif sample_data_schema is None and self.x_train is not None:
            sample_data_schema = self.x_train
            logger.warning('No dataset is provided, use the data schema from '
                           'training dataset(x_train) as the default one.')

        if type(lst_stats_name) is not list:
            raise FLException("list of requested statistics badly form. "
                              "It should of type list, but it is instead "
                              "of type " + str(type(lst_stats_name)))
        list_stats = {}
        if not isinstance(sample_data_schema, numpy.ndarray):
            raise FLException("Expecting the local dataset to be of type "
                              "numpy.ndarray, instead it is of type " +
                              str(type(sample_data_schema)))

        while len(lst_stats_name) != 0:
            tmp_stat_name = lst_stats_name.pop()
            if type(tmp_stat_name) is not str:
                logger.warning("Skipping the current requested statistics "
                               "name. It should be of type string, "
                               "but now it is instead of type" +
                               str(type(tmp_stat_name)))
            elif tmp_stat_name == 'min' or tmp_stat_name == 'minimum':
                list_stats[tmp_stat_name] = get_min(sample_data_schema)
            elif tmp_stat_name == 'max' or tmp_stat_name == 'maximum':
                list_stats[tmp_stat_name] = get_max(sample_data_schema)
            elif tmp_stat_name == 'mean':
                list_stats[tmp_stat_name] = get_mean(sample_data_schema)
            elif tmp_stat_name == 'var' or tmp_stat_name == 'variance':
                list_stats[tmp_stat_name] = get_var(sample_data_schema)
            elif tmp_stat_name == 'std' or \
                    tmp_stat_name == 'standard deviation':
                list_stats[tmp_stat_name] = get_std(sample_data_schema)
            elif tmp_stat_name == 'quantile':
                if kwargs and 'q' in kwargs:
                    list_stats[tmp_stat_name] = get_quantile(
                        sample_data_schema,
                        percentage=kwargs['q'])
                else:
                    raise FLException('Cannot compute quantile, '
                                      'missing quantile requirement.')
            else:
                logger.warning("Current required statistics "
                               "is not supported. Skipping...")

        return list_stats

    def get_preprocessor(self, sample_data_schema, preprocessor_name, **kwargs):
        """
        Set the data preprocessor of the data handler class as the requested
        type of preprocessor. The supported preprocessors
        include `normalizer`, `standardscaler` and `minmaxscaler`.
        All provided based on `sklearn.preprocessing` module.
        The preprocessor can be applied to perform the
        required preprocessing step for the party's local dataset
        via `transform` method.

        :param sample_data_schema: Provided data with only feature values to \
        initialize the preprocessor. \
        Assuming the dataset has shape (num_samples, num_features).
        :type sample_data_schema: `np.array`
        :param preprocessor_name: The requested preprocessor name in lowercase.
        :type preprocessor_name: `str`
        :param kwargs: Additional parameters to obtain the preprocessor.
        :type kwargs: `dict`
        :return: None
        """
        if sample_data_schema is None and self.x_train is None:
            raise FLException("No data is provided!")
        elif sample_data_schema is None and self.x_train is not None:
            sample_data_schema = self.x_train
            logger.warning('No dataset is provided, use the data schema from '
                           'training dataset(x_train) as the default one.')

        if type(preprocessor_name) is not str:
            raise FLException("Expecting the requested preprocessor "
                              "to be of type string, instead it is of type" +
                              str(type(preprocessor_name)))

        if not isinstance(sample_data_schema, numpy.ndarray):
            raise FLException("Expecting the local dataset to be of type "
                              "numpy.ndarray, instead it is of type " +
                              str(type(sample_data_schema)))

        if preprocessor_name == 'normalizer' or \
                preprocessor_name == 'normalization':
            if 'norm' in kwargs:
                self.preprocessor = get_normalizer(sample_data_schema,
                                                   norm=kwargs['norm'])
            else:
                self.preprocessor = get_normalizer(sample_data_schema)
        elif preprocessor_name == 'standardscaler' or \
                preprocessor_name == 'standardization':
            mean_val = None
            std = None
            if 'mean' in kwargs:
                mean_val = kwargs['mean']
            if 'scale' in kwargs:
                std = kwargs['scale']
            self.preprocessor = get_standardscaler(sample_data_schema,
                                                   mean_val=mean_val,
                                                   std=std)
        elif preprocessor_name == 'minmaxscaler':
            if 'feature_range' in kwargs:
                self.preprocessor = get_minmaxscaler(
                    sample_data_schema,
                    feature_range=kwargs['feature_range'])
            else:
                self.preprocessor = get_minmaxscaler(sample_data_schema)
        else:
            logger.warning("Required preprocessor is not supported. "
                           "Skipping...")
